
package org.mule.modules.vertex.config;

import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.lang.reflect.WildcardType;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.concurrent.atomic.AtomicInteger;
import org.mule.api.MessagingException;
import org.mule.api.MuleContext;
import org.mule.api.MuleEvent;
import org.mule.api.MuleException;
import org.mule.api.MuleMessage;
import org.mule.api.construct.FlowConstruct;
import org.mule.api.construct.FlowConstructAware;
import org.mule.api.context.MuleContextAware;
import org.mule.api.expression.ExpressionManager;
import org.mule.api.lifecycle.Disposable;
import org.mule.api.lifecycle.Initialisable;
import org.mule.api.lifecycle.InitialisationException;
import org.mule.api.lifecycle.Startable;
import org.mule.api.lifecycle.Stoppable;
import org.mule.api.processor.MessageProcessor;
import org.mule.api.registry.RegistrationException;
import org.mule.api.transformer.DataType;
import org.mule.api.transformer.Transformer;
import org.mule.api.transformer.TransformerException;
import org.mule.config.i18n.CoreMessages;
import org.mule.config.i18n.MessageFactory;
import org.mule.modules.vertex.VertexModule;
import org.mule.transformer.TransformerTemplate;
import org.mule.transformer.types.DataTypeFactory;
import org.mule.transport.NullPayload;
import org.mule.util.TemplateParser;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


/**
 * CalculateAccrualSyncTaxMessageProcessor invokes the {@link org.mule.modules.vertex.VertexModule#calculateAccrualSyncTax(java.util.Map, java.util.Map, java.util.Map, java.util.Map, java.util.Map, java.util.List, java.util.List, java.util.Map, java.util.Map, java.util.List, java.lang.String, java.util.Date, java.lang.String, java.lang.Boolean, java.lang.Boolean, org.mule.modules.vertex.DeliveryTermCodeType, java.util.Date, java.lang.String, org.mule.modules.vertex.ProcurementTransactionType, java.util.Date)} method in {@link VertexModule }. For each argument there is a field in this processor to match it.  Before invoking the actual method the processor will evaluate and transform where possible to the expected argument type.
 * 
 */
public class CalculateAccrualSyncTaxMessageProcessor
    implements FlowConstructAware, MuleContextAware, Disposable, Initialisable, Startable, Stoppable, MessageProcessor
{

    private Object currency;
    private Map<String, Object> _currencyType;
    private Object originalCurrency;
    private Map<String, Object> _originalCurrencyType;
    private Object buyer;
    private Map<String, Object> _buyerType;
    private Object vendor;
    private Map<String, Object> _vendorType;
    private Object taxOverride;
    private Map<String, Object> _taxOverrideType;
    private Object impositionsToProcess;
    private List<Map<String, Object>> _impositionsToProcessType;
    private Object jurisdictionOverrides;
    private List<Map<String, Object>> _jurisdictionOverridesType;
    private Object situsOverride;
    private Map<String, Object> _situsOverrideType;
    private Object chargedTax;
    private Map<String, Object> _chargedTaxType;
    private Object lineItems;
    private List<Map<String, Object>> _lineItemsType;
    private Object documentNumber;
    private String _documentNumberType;
    private Object postingDate;
    private Date _postingDateType;
    private Object locationCode;
    private String _locationCodeType;
    private Object returnAssistedParametersIndicator;
    private Boolean _returnAssistedParametersIndicatorType;
    private Object returnGeneratedLineItemsIndicator;
    private Boolean _returnGeneratedLineItemsIndicatorType;
    private Object deliveryTerm;
    private org.mule.modules.vertex.DeliveryTermCodeType _deliveryTermType;
    private Object documentDate;
    private Date _documentDateType;
    private Object transactionId;
    private String _transactionIdType;
    private Object transactionType;
    private org.mule.modules.vertex.ProcurementTransactionType _transactionTypeType;
    private Object recoverableDate;
    private Date _recoverableDateType;
    private static Logger logger = LoggerFactory.getLogger(CalculateAccrualSyncTaxMessageProcessor.class);
    /**
     * Module object
     * 
     */
    private Object moduleObject;
    /**
     * Mule Context
     * 
     */
    private MuleContext muleContext;
    /**
     * Mule Expression Manager
     * 
     */
    private ExpressionManager expressionManager;
    /**
     * Mule Pattern Info
     * 
     */
    private TemplateParser.PatternInfo patternInfo;
    /**
     * Flow construct
     * 
     */
    private FlowConstruct flowConstruct;
    /**
     * Variable used to track how many retries we have attempted on this message processor
     * 
     */
    private AtomicInteger retryCount;
    /**
     * Maximum number of retries that can be attempted.
     * 
     */
    private int retryMax;

    /**
     * Obtains the expression manager from the Mule context and initialises the connector. If a target object  has not been set already it will search the Mule registry for a default one.
     * 
     * @throws InitialisationException
     */
    public void initialise()
        throws InitialisationException
    {
        retryCount = new AtomicInteger();
        expressionManager = muleContext.getExpressionManager();
        patternInfo = TemplateParser.createMuleStyleParser().getStyle();
        if (moduleObject == null) {
            try {
                moduleObject = muleContext.getRegistry().lookupObject(VertexModuleLifecycleAdapter.class);
                if (moduleObject == null) {
                    throw new InitialisationException(MessageFactory.createStaticMessage("Cannot find object"), this);
                }
            } catch (RegistrationException e) {
                throw new InitialisationException(CoreMessages.initialisationFailure("org.mule.modules.vertex.config.VertexModuleLifecycleAdapter"), e, this);
            }
        }
        if (moduleObject instanceof String) {
            moduleObject = muleContext.getRegistry().lookupObject(((String) moduleObject));
            if (moduleObject == null) {
                throw new InitialisationException(MessageFactory.createStaticMessage("Cannot find object by config name"), this);
            }
        }
    }

    public void start()
        throws MuleException
    {
    }

    public void stop()
        throws MuleException
    {
    }

    public void dispose() {
    }

    /**
     * Set the Mule context
     * 
     * @param context Mule context to set
     */
    public void setMuleContext(MuleContext context) {
        this.muleContext = context;
    }

    /**
     * Sets flow construct
     * 
     * @param flowConstruct Flow construct to set
     */
    public void setFlowConstruct(FlowConstruct flowConstruct) {
        this.flowConstruct = flowConstruct;
    }

    /**
     * Sets the instance of the object under which the processor will execute
     * 
     * @param moduleObject Instace of the module
     */
    public void setModuleObject(Object moduleObject) {
        this.moduleObject = moduleObject;
    }

    /**
     * Sets retryMax
     * 
     * @param value Value to set
     */
    public void setRetryMax(int value) {
        this.retryMax = value;
    }

    /**
     * Sets documentDate
     * 
     * @param value Value to set
     */
    public void setDocumentDate(Object value) {
        this.documentDate = value;
    }

    /**
     * Sets postingDate
     * 
     * @param value Value to set
     */
    public void setPostingDate(Object value) {
        this.postingDate = value;
    }

    /**
     * Sets impositionsToProcess
     * 
     * @param value Value to set
     */
    public void setImpositionsToProcess(Object value) {
        this.impositionsToProcess = value;
    }

    /**
     * Sets jurisdictionOverrides
     * 
     * @param value Value to set
     */
    public void setJurisdictionOverrides(Object value) {
        this.jurisdictionOverrides = value;
    }

    /**
     * Sets transactionId
     * 
     * @param value Value to set
     */
    public void setTransactionId(Object value) {
        this.transactionId = value;
    }

    /**
     * Sets situsOverride
     * 
     * @param value Value to set
     */
    public void setSitusOverride(Object value) {
        this.situsOverride = value;
    }

    /**
     * Sets taxOverride
     * 
     * @param value Value to set
     */
    public void setTaxOverride(Object value) {
        this.taxOverride = value;
    }

    /**
     * Sets vendor
     * 
     * @param value Value to set
     */
    public void setVendor(Object value) {
        this.vendor = value;
    }

    /**
     * Sets returnGeneratedLineItemsIndicator
     * 
     * @param value Value to set
     */
    public void setReturnGeneratedLineItemsIndicator(Object value) {
        this.returnGeneratedLineItemsIndicator = value;
    }

    /**
     * Sets recoverableDate
     * 
     * @param value Value to set
     */
    public void setRecoverableDate(Object value) {
        this.recoverableDate = value;
    }

    /**
     * Sets currency
     * 
     * @param value Value to set
     */
    public void setCurrency(Object value) {
        this.currency = value;
    }

    /**
     * Sets deliveryTerm
     * 
     * @param value Value to set
     */
    public void setDeliveryTerm(Object value) {
        this.deliveryTerm = value;
    }

    /**
     * Sets returnAssistedParametersIndicator
     * 
     * @param value Value to set
     */
    public void setReturnAssistedParametersIndicator(Object value) {
        this.returnAssistedParametersIndicator = value;
    }

    /**
     * Sets originalCurrency
     * 
     * @param value Value to set
     */
    public void setOriginalCurrency(Object value) {
        this.originalCurrency = value;
    }

    /**
     * Sets locationCode
     * 
     * @param value Value to set
     */
    public void setLocationCode(Object value) {
        this.locationCode = value;
    }

    /**
     * Sets transactionType
     * 
     * @param value Value to set
     */
    public void setTransactionType(Object value) {
        this.transactionType = value;
    }

    /**
     * Sets buyer
     * 
     * @param value Value to set
     */
    public void setBuyer(Object value) {
        this.buyer = value;
    }

    /**
     * Sets chargedTax
     * 
     * @param value Value to set
     */
    public void setChargedTax(Object value) {
        this.chargedTax = value;
    }

    /**
     * Sets documentNumber
     * 
     * @param value Value to set
     */
    public void setDocumentNumber(Object value) {
        this.documentNumber = value;
    }

    /**
     * Sets lineItems
     * 
     * @param value Value to set
     */
    public void setLineItems(Object value) {
        this.lineItems = value;
    }

    /**
     * Get all superclasses and interfaces recursively.
     * 
     * @param classes List of classes to which to add all found super classes and interfaces.
     * @param clazz   The class to start the search with.
     */
    private void computeClassHierarchy(Class clazz, List classes) {
        for (Class current = clazz; (current!= null); current = current.getSuperclass()) {
            if (classes.contains(current)) {
                return ;
            }
            classes.add(current);
            for (Class currentInterface: current.getInterfaces()) {
                computeClassHierarchy(currentInterface, classes);
            }
        }
    }

    /**
     * Checks whether the specified class parameter is an instance of {@link List }
     * 
     * @param clazz <code>Class</code> to check.
     * @return
     */
    private boolean isListClass(Class clazz) {
        List<Class> classes = new ArrayList<Class>();
        computeClassHierarchy(clazz, classes);
        return classes.contains(List.class);
    }

    /**
     * Checks whether the specified class parameter is an instance of {@link Map }
     * 
     * @param clazz <code>Class</code> to check.
     * @return
     */
    private boolean isMapClass(Class clazz) {
        List<Class> classes = new ArrayList<Class>();
        computeClassHierarchy(clazz, classes);
        return classes.contains(Map.class);
    }

    private boolean isList(Type type) {
        if ((type instanceof Class)&&isListClass(((Class) type))) {
            return true;
        }
        if (type instanceof ParameterizedType) {
            return isList(((ParameterizedType) type).getRawType());
        }
        if (type instanceof WildcardType) {
            Type[] upperBounds = ((WildcardType) type).getUpperBounds();
            return ((upperBounds.length!= 0)&&isList(upperBounds[ 0 ]));
        }
        return false;
    }

    private boolean isMap(Type type) {
        if ((type instanceof Class)&&isMapClass(((Class) type))) {
            return true;
        }
        if (type instanceof ParameterizedType) {
            return isMap(((ParameterizedType) type).getRawType());
        }
        if (type instanceof WildcardType) {
            Type[] upperBounds = ((WildcardType) type).getUpperBounds();
            return ((upperBounds.length!= 0)&&isMap(upperBounds[ 0 ]));
        }
        return false;
    }

    private boolean isAssignableFrom(Type expectedType, Class clazz) {
        if (expectedType instanceof Class) {
            return ((Class) expectedType).isAssignableFrom(clazz);
        }
        if (expectedType instanceof ParameterizedType) {
            return isAssignableFrom(((ParameterizedType) expectedType).getRawType(), clazz);
        }
        if (expectedType instanceof WildcardType) {
            Type[] upperBounds = ((WildcardType) expectedType).getUpperBounds();
            if (upperBounds.length!= 0) {
                return isAssignableFrom(upperBounds[ 0 ], clazz);
            }
        }
        return false;
    }

    private Object evaluate(MuleMessage muleMessage, Object source) {
        if (source instanceof String) {
            String stringSource = ((String) source);
            if (stringSource.startsWith(patternInfo.getPrefix())&&stringSource.endsWith(patternInfo.getSuffix())) {
                return expressionManager.evaluate(stringSource, muleMessage);
            } else {
                return expressionManager.parse(stringSource, muleMessage);
            }
        }
        return source;
    }

    private Object evaluateAndTransform(MuleMessage muleMessage, Type expectedType, String expectedMimeType, Object source)
        throws TransformerException
    {
        if (source == null) {
            return source;
        }
        Object target = null;
        if (isList(source.getClass())) {
            if (isList(expectedType)) {
                List newList = new ArrayList();
                Type valueType = ((ParameterizedType) expectedType).getActualTypeArguments()[ 0 ];
                ListIterator iterator = ((List) source).listIterator();
                while (iterator.hasNext()) {
                    Object subTarget = iterator.next();
                    newList.add(evaluateAndTransform(muleMessage, valueType, expectedMimeType, subTarget));
                }
                target = newList;
            } else {
                target = source;
            }
        } else {
            if (isMap(source.getClass())) {
                if (isMap(expectedType)) {
                    Type keyType = Object.class;
                    Type valueType = Object.class;
                    if (expectedType instanceof ParameterizedType) {
                        keyType = ((ParameterizedType) expectedType).getActualTypeArguments()[ 0 ];
                        valueType = ((ParameterizedType) expectedType).getActualTypeArguments()[ 1 ];
                    }
                    Map map = ((Map) source);
                    Map newMap = new HashMap();
                    for (Object entryObj: map.entrySet()) {
                        {
                            Map.Entry entry = ((Map.Entry) entryObj);
                            Object newKey = evaluateAndTransform(muleMessage, keyType, expectedMimeType, entry.getKey());
                            Object newValue = evaluateAndTransform(muleMessage, valueType, expectedMimeType, entry.getValue());
                            newMap.put(newKey, newValue);
                        }
                    }
                    target = newMap;
                } else {
                    target = source;
                }
            } else {
                target = evaluate(muleMessage, source);
            }
        }
        if ((target!= null)&&(!isAssignableFrom(expectedType, target.getClass()))) {
            DataType sourceDataType = DataTypeFactory.create(target.getClass());
            DataType targetDataType = null;
            if (expectedMimeType!= null) {
                targetDataType = DataTypeFactory.create(((Class) expectedType), expectedMimeType);
            } else {
                targetDataType = DataTypeFactory.create(((Class) expectedType));
            }
            Transformer t = muleContext.getRegistry().lookupTransformer(sourceDataType, targetDataType);
            return t.transform(target);
        } else {
            return target;
        }
    }

    /**
     * Invokes the MessageProcessor.
     * 
     * @param event MuleEvent to be processed
     * @throws MuleException
     */
    public MuleEvent process(MuleEvent event)
        throws MuleException
    {
        MuleMessage _muleMessage = event.getMessage();
        VertexModuleLifecycleAdapter castedModuleObject = null;
        if (moduleObject instanceof String) {
            castedModuleObject = ((VertexModuleLifecycleAdapter) muleContext.getRegistry().lookupObject(((String) moduleObject)));
            if (castedModuleObject == null) {
                throw new MessagingException(CoreMessages.failedToCreate("calculateAccrualSyncTax"), event, new RuntimeException("Cannot find the configuration specified by the config-ref attribute."));
            }
        } else {
            castedModuleObject = ((VertexModuleLifecycleAdapter) moduleObject);
        }
        try {
            Map<String, Object> transformedCurrency = ((Map<String, Object> ) evaluateAndTransform(_muleMessage, CalculateAccrualSyncTaxMessageProcessor.class.getDeclaredField("_currencyType").getGenericType(), null, currency));
            Map<String, Object> transformedOriginalCurrency = ((Map<String, Object> ) evaluateAndTransform(_muleMessage, CalculateAccrualSyncTaxMessageProcessor.class.getDeclaredField("_originalCurrencyType").getGenericType(), null, originalCurrency));
            Map<String, Object> transformedBuyer = ((Map<String, Object> ) evaluateAndTransform(_muleMessage, CalculateAccrualSyncTaxMessageProcessor.class.getDeclaredField("_buyerType").getGenericType(), null, buyer));
            Map<String, Object> transformedVendor = ((Map<String, Object> ) evaluateAndTransform(_muleMessage, CalculateAccrualSyncTaxMessageProcessor.class.getDeclaredField("_vendorType").getGenericType(), null, vendor));
            Map<String, Object> transformedTaxOverride = ((Map<String, Object> ) evaluateAndTransform(_muleMessage, CalculateAccrualSyncTaxMessageProcessor.class.getDeclaredField("_taxOverrideType").getGenericType(), null, taxOverride));
            List<Map<String, Object>> transformedImpositionsToProcess = ((List<Map<String, Object>> ) evaluateAndTransform(_muleMessage, CalculateAccrualSyncTaxMessageProcessor.class.getDeclaredField("_impositionsToProcessType").getGenericType(), null, impositionsToProcess));
            List<Map<String, Object>> transformedJurisdictionOverrides = ((List<Map<String, Object>> ) evaluateAndTransform(_muleMessage, CalculateAccrualSyncTaxMessageProcessor.class.getDeclaredField("_jurisdictionOverridesType").getGenericType(), null, jurisdictionOverrides));
            Map<String, Object> transformedSitusOverride = ((Map<String, Object> ) evaluateAndTransform(_muleMessage, CalculateAccrualSyncTaxMessageProcessor.class.getDeclaredField("_situsOverrideType").getGenericType(), null, situsOverride));
            Map<String, Object> transformedChargedTax = ((Map<String, Object> ) evaluateAndTransform(_muleMessage, CalculateAccrualSyncTaxMessageProcessor.class.getDeclaredField("_chargedTaxType").getGenericType(), null, chargedTax));
            List<Map<String, Object>> transformedLineItems = ((List<Map<String, Object>> ) evaluateAndTransform(_muleMessage, CalculateAccrualSyncTaxMessageProcessor.class.getDeclaredField("_lineItemsType").getGenericType(), null, lineItems));
            String transformedDocumentNumber = ((String) evaluateAndTransform(_muleMessage, CalculateAccrualSyncTaxMessageProcessor.class.getDeclaredField("_documentNumberType").getGenericType(), null, documentNumber));
            Date transformedPostingDate = ((Date) evaluateAndTransform(_muleMessage, CalculateAccrualSyncTaxMessageProcessor.class.getDeclaredField("_postingDateType").getGenericType(), null, postingDate));
            String transformedLocationCode = ((String) evaluateAndTransform(_muleMessage, CalculateAccrualSyncTaxMessageProcessor.class.getDeclaredField("_locationCodeType").getGenericType(), null, locationCode));
            Boolean transformedReturnAssistedParametersIndicator = ((Boolean) evaluateAndTransform(_muleMessage, CalculateAccrualSyncTaxMessageProcessor.class.getDeclaredField("_returnAssistedParametersIndicatorType").getGenericType(), null, returnAssistedParametersIndicator));
            Boolean transformedReturnGeneratedLineItemsIndicator = ((Boolean) evaluateAndTransform(_muleMessage, CalculateAccrualSyncTaxMessageProcessor.class.getDeclaredField("_returnGeneratedLineItemsIndicatorType").getGenericType(), null, returnGeneratedLineItemsIndicator));
            org.mule.modules.vertex.DeliveryTermCodeType transformedDeliveryTerm = ((org.mule.modules.vertex.DeliveryTermCodeType) evaluateAndTransform(_muleMessage, CalculateAccrualSyncTaxMessageProcessor.class.getDeclaredField("_deliveryTermType").getGenericType(), null, deliveryTerm));
            Date transformedDocumentDate = ((Date) evaluateAndTransform(_muleMessage, CalculateAccrualSyncTaxMessageProcessor.class.getDeclaredField("_documentDateType").getGenericType(), null, documentDate));
            String transformedTransactionId = ((String) evaluateAndTransform(_muleMessage, CalculateAccrualSyncTaxMessageProcessor.class.getDeclaredField("_transactionIdType").getGenericType(), null, transactionId));
            org.mule.modules.vertex.ProcurementTransactionType transformedTransactionType = ((org.mule.modules.vertex.ProcurementTransactionType) evaluateAndTransform(_muleMessage, CalculateAccrualSyncTaxMessageProcessor.class.getDeclaredField("_transactionTypeType").getGenericType(), null, transactionType));
            Date transformedRecoverableDate = ((Date) evaluateAndTransform(_muleMessage, CalculateAccrualSyncTaxMessageProcessor.class.getDeclaredField("_recoverableDateType").getGenericType(), null, recoverableDate));
            retryCount.getAndIncrement();
            Object resultPayload;
            resultPayload = castedModuleObject.calculateAccrualSyncTax(transformedCurrency, transformedOriginalCurrency, transformedBuyer, transformedVendor, transformedTaxOverride, transformedImpositionsToProcess, transformedJurisdictionOverrides, transformedSitusOverride, transformedChargedTax, transformedLineItems, transformedDocumentNumber, transformedPostingDate, transformedLocationCode, transformedReturnAssistedParametersIndicator, transformedReturnGeneratedLineItemsIndicator, transformedDeliveryTerm, transformedDocumentDate, transformedTransactionId, transformedTransactionType, transformedRecoverableDate);
            TransformerTemplate.OverwitePayloadCallback overwritePayloadCallback = null;
            if (resultPayload == null) {
                overwritePayloadCallback = new TransformerTemplate.OverwitePayloadCallback(NullPayload.getInstance());
            } else {
                overwritePayloadCallback = new TransformerTemplate.OverwitePayloadCallback(resultPayload);
            }
            List<Transformer> transformerList;
            transformerList = new ArrayList<Transformer>();
            transformerList.add(new TransformerTemplate(overwritePayloadCallback));
            event.getMessage().applyTransformers(event, transformerList);
            retryCount.set(0);
            return event;
        } catch (Exception e) {
            throw new MessagingException(CoreMessages.failedToInvoke("calculateAccrualSyncTax"), event, e);
        }
    }

}
